#include <vector>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <msp/core/systemerror.h>
#include "display_private.h"
#include "window.h"
#include "window_private.h"

using namespace std;

namespace {

Bool match_event_type(Display *, XEvent *event, XPointer arg)
{
	return event->type==*reinterpret_cast<int *>(arg);
}

}

namespace Msp {
namespace Graphics {

void Window::platform_init()
{
	DisplayHandle dpy = display.get_private().display;

	priv->wm_delete_window = XInternAtom(dpy, "WM_DELETE_WINDOW", true);
	priv->invisible_cursor = 0;
	priv->reparented = false;
	priv->rel_x = 0;
	priv->rel_y = 0;

	XSetWindowAttributes attr;
	attr.override_redirect = options.fullscreen;
	attr.event_mask = ButtonPressMask|ButtonReleaseMask|PointerMotionMask|KeyPressMask|KeyReleaseMask|StructureNotifyMask|EnterWindowMask|ExposureMask;

	priv->window = XCreateWindow(dpy,
		display.get_private().root_window,
		0, 0,    // User position is set when the window is mapped
		options.width, options.height,
		0,
		CopyFromParent,
		InputOutput,
		CopyFromParent,
		CWOverrideRedirect|CWEventMask, &attr);

	XSetWMProtocols(dpy, priv->window, &priv->wm_delete_window, 1);

	if(!options.resizable)
	{
		XSizeHints hints;
		hints.flags = PMinSize|PMaxSize;
		hints.min_width=hints.max_width = options.width;
		hints.min_height=hints.max_height = options.height;
		XSetWMNormalHints(dpy, priv->window, &hints);
	}
}

void Window::platform_cleanup()
{
	if(priv->window)
		XDestroyWindow(display.get_private().display, priv->window);

	if(priv->invisible_cursor)
		XFreeCursor(display.get_private().display, priv->invisible_cursor);
}

void Window::set_title(const string &title)
{
	vector<unsigned char> buf(title.begin(), title.end());
	XTextProperty prop;
	prop.value = &buf[0];
	prop.encoding = XA_STRING;
	prop.format = 8;
	prop.nitems = title.size();
	XSetWMName(display.get_private().display, priv->window, &prop);
	display.check_error();
}

void Window::platform_reconfigure(bool fullscreen_changed)
{
	DisplayHandle dpy = display.get_private().display;

	bool was_visible = visible;
	if(fullscreen_changed)
	{
		if(was_visible)
		{
			hide();

			// Wait for the window to be unmapped.  This makes window managers happy.
			XEvent ev;
			int ev_type = UnmapNotify;
			XPeekIfEvent(dpy, &ev, match_event_type, reinterpret_cast<char *>(&ev_type));
		}

		XSetWindowAttributes attr;
		attr.override_redirect = options.fullscreen;
		XChangeWindowAttributes(dpy, priv->window, CWOverrideRedirect, &attr);
	}

	XSizeHints hints;
	if(options.resizable)
		hints.flags = 0;
	else
	{
		hints.flags = PMinSize|PMaxSize;
		hints.min_width = hints.max_width = options.width;
		hints.min_height = hints.max_height = options.height;
	}
	XSetWMNormalHints(dpy, priv->window, &hints);

	if(options.fullscreen)
		XMoveResizeWindow(dpy, priv->window, 0, 0, options.width, options.height);
	else if(options.user_position)
		XMoveResizeWindow(dpy, priv->window, options.x, options.y, options.width, options.height);
	else
		XResizeWindow(dpy, priv->window, options.width, options.height);

	if(fullscreen_changed && was_visible)
		show();
}

void Window::show_cursor(bool s)
{
	DisplayHandle dpy = display.get_private().display;

	if(s)
		XUndefineCursor(dpy, priv->window);
	else
	{
		if(!priv->invisible_cursor)
		{
			int screen = DefaultScreen(dpy);

			Pixmap pm = XCreatePixmap(dpy, priv->window, 1, 1, 1);
			GC gc = XCreateGC(dpy, pm, 0, 0);
			XSetFunction(dpy, gc, GXclear);
			XDrawPoint(dpy, pm, gc, 0, 0);
			XFreeGC(dpy, gc);

			XColor black;
			black.pixel = BlackPixel(dpy, screen);
			XQueryColor(dpy, DefaultColormap(dpy, screen), &black);

			priv->invisible_cursor = XCreatePixmapCursor(dpy, pm, pm, &black, &black, 0, 0);

			XFreePixmap(dpy, pm);
		}
		XDefineCursor(dpy, priv->window, priv->invisible_cursor);
	}
}

void Window::warp_pointer(int x, int y)
{
	XWarpPointer(display.get_private().display, None, priv->window, 0, 0, 0, 0, x, y);
}

void Window::platform_set_touch_input()
{
}

void Window::platform_show()
{
	DisplayHandle dpy = display.get_private().display;
	XMapRaised(dpy, priv->window);
	if(options.user_position)
		XMoveWindow(dpy, priv->window, options.x, options.y);
}

void Window::platform_hide()
{
	XUnmapWindow(display.get_private().display, priv->window);
}

bool Window::event(const Event &evnt)
{
	const XEvent &ev = evnt.xevent;
	switch(ev.type)
	{
	case ButtonPress:
	case ButtonRelease:
	case MotionNotify:
	case KeyPress:
	case KeyRelease:
		signal_input_event.emit(evnt);
		break;
	case ConfigureNotify:
		if((ev.xconfigure.width==static_cast<int>(options.width) && ev.xconfigure.height==static_cast<int>(options.height)) == resizing)
		{
			options.width = ev.xconfigure.width;
			options.height = ev.xconfigure.height;
			resizing = false;
			signal_resize.emit(options.width, options.height);
		}

		{
			int x = ev.xconfigure.x;
			int y = ev.xconfigure.y;

			if(priv->reparented)
			{
				if(!ev.xconfigure.send_event)
				{
					/* If the window manager reparented us, the coordinates of a
					real event are in the parent window's space. */
					priv->rel_x = ev.xconfigure.x;
					priv->rel_y = ev.xconfigure.y;

					const Display::Private &dpy_priv = display.get_private();
					WindowHandle dummy;
					XTranslateCoordinates(dpy_priv.display, priv->window, dpy_priv.root_window, 0, 0, &x, &y, &dummy);
				}

				// Use the coordinates of the window manager frame
				x -= priv->rel_x;
				y -= priv->rel_y;
			}

			if((x==options.x && y==options.y) == moving)
			{
				options.x = x;
				options.y = y;
				moving = false;
				signal_move.emit(options.x, options.y);
			}
		}

		break;
	case ReparentNotify:
		priv->reparented = (ev.xreparent.parent!=display.get_private().root_window);
		if(!priv->reparented)
		{
			priv->rel_x = 0;
			priv->rel_y = 0;
		}
		break;
	case ClientMessage:
		if(ev.xclient.data.l[0]==static_cast<long>(priv->wm_delete_window))
			signal_close.emit();
		break;
	case EnterNotify:
		if(options.fullscreen)
			XSetInputFocus(display.get_private().display, priv->window, RevertToParent, CurrentTime);
		break;
	case MapNotify:
		if(options.fullscreen)
			XGrabPointer(display.get_private().display, priv->window, true, None, GrabModeAsync, GrabModeAsync, priv->window, None, CurrentTime);
		break;
	case Expose:
		signal_expose.emit(ev.xexpose.x, ev.xexpose.y, ev.xexpose.width, ev.xexpose.height, evnt);
		break;
	default:
		return false;
	}

	return true;
}

} // namespace Graphics
} // namespace Msp
